/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PRISM_HEADER_HH_
# define PRISM_HEADER_HH_

# include <wipal/phy/phy.hh>

/// Prism %header related compounds.
namespace prism
{
  /**
   * @link header Prism %header @endlink %item.
   *
   * @see prism::header
   */
  struct item
  {
    /// Read the item's value according to given endianness.
    uint32_t	get(bool swapped) const;

    /// Set the item's value according to given endianness.
    item&	set(uint32_t value, bool swapped);

  protected:
    uint32_t did;	///< What this item's value is.
    uint16_t status;	///< Is this item's value present (0) or absent (1).
    uint16_t len;	///< Length in bytes, should be 4.
    uint32_t data;	///< Item value or zero.
  };

  /// Prism %header format.
  struct header
  {
    enum
      {
	/// @link header::devname @c devname @endlink's maximum length.
	devnamelen_max = 16
      };

    uint32_t	msgcode;			///< Message code.
    uint32_t	msglen;				///< Header length.
    uint8_t	devname[devnamelen_max];	///< Interface name.
    item	hosttime;			///< Host time.
    item	mactime;			///< MAC time.
    item	channel;			///< Channel.
    item	rssi;				///< RSSI.
    item	sq;				///< Signal quality(?).
    item	signal;				///< Signal quality(?).
    item	noise;				///< Noise information.
    item	rate;				///< Tx/Rx rate.
    item	istx;				///< Is TX.
    item	frmlen;				///< Frame length.
  };

  /**
   * Compare IEEE 802.11 bytes from Prism headers.
   *
   * Return whether two IEEE 802.11 frames embedded inside Prism
   * headers are equals.  This function <b>does not compare</b> the
   * Prism headers.
   *
   * In case there is not enough bytes to decide whether frames are
   * equal (caplen < sizeof prism::header), consider they are not.
   * This is a conservative approach: when merging two traces, this
   * will lead to both frames having their entry in the merged trace.
   */
  bool
  eq_80211(const header* lhs, size_t lhs_caplen,
	   const header* rhs, size_t rhs_caplen);

  /**
   * Compare IEEE 802.11 bytes from Prism headers, but also check
   * MAC time and host time.
   *
   * Return whether two IEEE 802.11 frames embedded inside Prism
   * headers are equals.  Also compare the MAC time fields of Prism
   * headers, and the host time fields if @c check_hosttime is true.
   * When the time difference between each header's field is less than
   * @c precision, then consider the field equals.
   *
   * In case there is not enough bytes to decide whether frames are
   * equal (caplen < sizeof prism::header), consider they are not.
   * This is a conservative approach: when merging two traces, this
   * will lead to both frames having their entry in the merged trace.
   */
  bool
  eq_time_and_80211(const header* lhs, size_t lhs_caplen, bool lhs_swapped,
		    const header* rhs, size_t rhs_caplen, bool rhs_swapped,
		    unsigned	precision = phy::eq_time_default_precision,
		    bool	check_hosttime = false);

} // End of namespace prism.

# include "prism_header.hxx"

#endif // ! PRISM_HEADER_HH_
