/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FILTER_WINDOWED_MERGE_HXX_
# define WIFI_FRAME_FILTER_WINDOWED_MERGE_HXX_

# include <iostream>

# include "windowed_merge.hh"

# include <wipal/wifi/frame/filter/merge.hh>

namespace wifi
{
  namespace frame
  {
    namespace filter
    {

      namespace internals
      {

# ifdef ENABLE_INFO
#  define wp_increment(X)	++(X)
#  define wp_increase(X, Y...)	((X) += (Y))
# else // ! ENABLE_INFO
#  define wp_increment(X)	(void) 0
#  define wp_increase(X, Y...)	(void) (Y)
# endif // ENABLE_INFO

	template <class HT, class I1, class I2, class B1, class B2>
	windowed_merge_iterator<HT, I1, I2, B1, B2>::
	windowed_merge_iterator(const iterable_type& i, bool end):
	  iterable_ (&i),
	  next1_ (end ? i.last1_ : i.first1_),
	  next2_ (end ? i.last2_ : i.first2_),
	  current_ (-1)
# ifdef ENABLE_INFO
	, lhs_total_ (0),
	  rhs_total_ (0),
	  shared_ (0),
	  total_ (0),
	  sync_error_ (0)
# endif // ENABLE_INFO
	{
	  if (not end)
	    {
	      wp_increase(lhs_total_, fill(win1_, next1_, iterable_->last1_,
					   window_size));
	      wp_increase(rhs_total_, fill(win2_, next2_, iterable_->last2_,
					   window_size));
	      current_ = first_packet();
	    }
	}

	template <class HT, class I1, class I2, class B1, class B2>
	bool
	windowed_merge_iterator<HT, I1, I2, B1, B2>::
	equal(const windowed_merge_iterator& rhs) const
	{
	  return (next1_ == rhs.next1_ and win1_.size() == rhs.win1_.size() and
		  next2_ == rhs.next2_ and win2_.size() == rhs.win2_.size());
	}

#define wp_advance(Counter, N)						\
	do								\
	  {								\
	    wp_increase(Counter, fill(win ## N ## _, next ## N ## _,	\
				      iterable_->last ## N ## _, 1));	\
	    win ## N ## _.pop_front();					\
	  }								\
	while (0)

	template <class HT, class I1, class I2, class B1, class B2>
	void
	windowed_merge_iterator<HT, I1, I2, B1, B2>::increment()
	{
	  assert(current_ == 0 or current_ == 1);

	  if (win1_.empty())
	    wp_advance(rhs_total_, 2);
	  else if (win2_.empty())
	    wp_advance(lhs_total_, 1);
	  else if (const optional_match r = match())
	    {
# ifdef ENABLE_INFO
	      {
		const tool::microseconds& t1 = r->first->microseconds();
		const tool::microseconds& t2 = r->second->microseconds();

		if (t1 > t2)
		  sync_error_ += mpz_class (t1 - t2).get_ui();
		else
		  sync_error_ += mpz_class (t2 - t1).get_ui();
	      }
	      ++shared_;
# endif // ENABLE_INFO
	      win1_.erase(r->first);
	      win2_.erase(r->second);
	      wp_increase(lhs_total_,
			  fill(win1_, next1_, iterable_->last1_, 1));
	      wp_increase(rhs_total_,
			  fill(win2_, next2_, iterable_->last2_, 1));
	    }
	  else
	    {
	      if (current_)
		wp_advance(rhs_total_, 2);
	      else
		wp_advance(lhs_total_, 1);
	    }
	  wp_increment(total_);

	  current_ = first_packet();

# ifdef ENABLE_INFO
	  if (current_ == -1)
	    std::cerr << "INFO: " << shared_ << " frames shared by"
		                                " both traces.\n"
		      << "INFO: Trace 1 has " << lhs_total_ << " frames ("
		      << (shared_ * 100.f / lhs_total_) << "% shared).\n"
		      << "INFO: Trace 2 has " << rhs_total_ << " frames ("
		      << (shared_ * 100.f / rhs_total_) << "% shared).\n"
		      << "INFO: Output has " << total_ << " frames ("
		      << (shared_ * 100.f / total_) << "% shared).\n"
		      << "INFO: Mean sync. error is "
		      << (sync_error_ / shared_) << " microseconds."
		      << std::endl;
# endif // ENABLE_INFO
	}

# undef wp_advance
# undef wp_increment
# undef wp_increase

	template <class HT, class I1, class I2, class B1, class B2>
	const typename windowed_merge_iterator<HT, I1, I2, B1, B2>::value_type&
	windowed_merge_iterator<HT, I1, I2, B1, B2>::get() const
	{
	  assert(current_ == 0 or current_ == 1);

	  if (current_)
	    return win2_.front();
	  else
	    return win1_.front();
	}

	template <class HT, class I1, class I2, class B1, class B2>
	typename windowed_merge_iterator<HT, I1, I2, B1, B2>::value_type&
	windowed_merge_iterator<HT, I1, I2, B1, B2>::get()
	{
	  assert(current_ == 0 or current_ == 1);

	  if (current_)
	    return win2_.front();
	  else
	    return win1_.front();
	}

	template <class HT, class I1, class I2, class B1, class B2>
	const typename windowed_merge_iterator<HT, I1, I2, B1, B2>::value_type*
	windowed_merge_iterator<HT, I1, I2, B1, B2>::get_ptr() const
	{
	  assert(current_ == 0 or current_ == 1);

	  if (current_)
	    return & win2_.front();
	  else
	    return & win1_.front();
	}

	template <class HT, class I1, class I2, class B1, class B2>
	typename windowed_merge_iterator<HT, I1, I2, B1, B2>::value_type*
	windowed_merge_iterator<HT, I1, I2, B1, B2>::get_ptr()
	{
	  assert(current_ == 0 or current_ == 1);

	  if (current_)
	    return & win2_.front();
	  else
	    return & win1_.front();
	}

	template <class HT, class I1, class I2, class B1, class B2>
	int
	windowed_merge_iterator<HT, I1, I2, B1, B2>::first_packet() const
	{
	  const bool first_at_end	= win1_.empty();
	  const bool second_at_end	= win2_.empty();

	  if (first_at_end)
	    return second_at_end ? -1 : 1;
	  if (second_at_end)
	    return 0;

	  return win1_.front().microseconds() < win2_.front().microseconds() ?
	    0 : 1;
	}

	template <class HT, class I1, class I2, class B1, class B2>
	template <class I>
	unsigned
	windowed_merge_iterator<HT, I1, I2, B1, B2>::
	fill(window_type&	w,
	     I&			i,
	     const I&		last,
	     size_t		n)
	{
	  unsigned r;

	  for (r = 0; r < n and i != last; ++r)
	    w.push_back(*i++);
	  return r;
	}

	template <class HT, class I1, class I2, class B1, class B2>
	typename windowed_merge_iterator<HT, I1, I2, B1, B2>::optional_match
	windowed_merge_iterator<HT, I1, I2, B1, B2>::match()
	{
	  typedef internals::equal<HT>	equal;

	  const value_type&	v = current_ ? win2_.front() : win1_.front();
	  const win_iterator	f = current_ ? win1_.begin() : win2_.begin();
	  const win_iterator	l = current_ ? win1_.end()  : win2_.end();
	  const win_iterator	r =
	    std::find(f, l, equal (v, iterable_->phy_end_));

	  if (r == l)
	    return boost::none_t ();

	  return current_?
	    match_type (r, win2_.begin()):
	    match_type (win1_.begin(), r);
	}

      } // End of namespace internals.

      template <class HeaderType,
		class InputIterator1,
		class InputIterator2,
		class Bottom>
      windowed_merge<HeaderType, InputIterator1, InputIterator2, Bottom>::
      windowed_merge(const InputIterator1& first1, const InputIterator1& last1,
		     const InputIterator2& first2, const InputIterator2& last2,
		     tool::endian::endianness	phy_end):
	phy_end_ (phy_end),
	first1_ (first1),
	last1_ (last1),
	first2_ (first2),
	last2_ (last2)
      {
      }

    } // End of namespace wifi::frame::filter.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

#endif // ! WIFI_FRAME_FILTER_WINDOWED_MERGE_HXX_
