/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_MACTIME_TRACKER_HH_
# define WIFI_MACTIME_TRACKER_HH_

extern "C"
{
# include <inttypes.h>
}

# include <wipal/tool/static_base.hh>
# include <wipal/tool/microseconds.hh>
# include <wipal/phy/prism_header.hh>

namespace wifi
{

  /**
   * Microsecond-grained time tracker.
   *
   * Prism headers have a @c mactime field that holds frames' MAC
   * arrival time.  But this field is only 32 bits long, thus wrapping
   * every one hour and a half (roughly).  This class tracks the
   * evolution of @c mactime fields from Prism headers and maintains
   * a microsecond count that does not wrap.
   *
   * Just call the tick() method for every @c mactime fields you get
   * in your trace.
   */
  template <class Bottom = tool::bottom>
  struct mactime_tracker: wp_inherit(public tool::static_base,
				     mactime_tracker<Bottom>)
  {
    /// Type of Prism headers' @c mactime fields.
    typedef uint32_t		mactime_type;

    /// Default constructor - initialize the microsecond count to zero.
    mactime_tracker();

    /// Construct a @c mactime tracker according to the given starting time.
    mactime_tracker(const mactime_type& mactime_start);

    /// Update the tracker with the giver @c mactime.
    void			tick(const mactime_type&);

    /// Update the tracker with the given Prism header's MAC time.
    void			tick(const prism::header*	header,
				     size_t			caplen,
				     bool			swapped);

    /// Retrieve the current microsecond count without wraps.
    tool::microseconds		microseconds() const;

    /// Retrieve the current mactime (might wrap).
    const mactime_type&		mactime() const;

  private:
    mactime_type	mactime_;
    unsigned		wrap_count_;
  };

} // End of namespace wifi.

# include "mactime_tracker.hxx"

#endif // ! WIFI_MACTIME_TRACKER_HH_
